@@ -9,7 +9,7 @@ except ImportError:
 
 from flask import Blueprint, render_template, abort, request, url_for, redirect, flash, Response
 from flask_peewee.filters import FilterMapping, FilterForm, FilterModelConverter
-from flask_peewee.forms import BaseModelConverter
+from flask_peewee.forms import BaseModelConverter, ChosenAjaxSelectWidget, LimitedModelSelectField
 from flask_peewee.serializer import Serializer
 from flask_peewee.utils import get_next, PaginatedQuery, path_to_models, slugify
 from peewee import BooleanField, DateTimeField, ForeignKeyField, DateField, TextField
@@ -27,8 +27,6 @@ class AdminModelConverter(BaseModelConverter):
         super(AdminModelConverter, self).__init__(additional)
         self.model_admin = model_admin
 
-        self.converters[ForeignKeyField] = self.handle_foreign_key
-
     def handle_foreign_key(self, model, field, **kwargs):
         if field.null:
             kwargs['allow_blank'] = True
@@ -40,6 +38,21 @@ class AdminModelConverter(BaseModelConverter):
         return field.name, form_field
 
 
+class AdminFilterModelConverter(FilterModelConverter):
+    def __init__(self, model_admin, additional=None):
+        super(AdminFilterModelConverter, self).__init__(additional)
+        self.model_admin = model_admin
+
+    def handle_foreign_key(self, model, field, **kwargs):
+        if field.name in (self.model_admin.foreign_key_lookups or ()):
+            data_source = url_for(self.model_admin.get_url_name('ajax_list'))
+            widget = ChosenAjaxSelectWidget(data_source, field.name)
+            form_field = LimitedModelSelectField(model=field.rel_model, widget=widget, **kwargs)
+        else:
+            form_field = ModelSelectField(model=field.rel_model, **kwargs)
+        return field.name, form_field
+
+
 class ModelAdmin(object):
     """
     ModelAdmin provides create/edit/delete functionality for a peewee Model.
@@ -70,7 +83,7 @@ class ModelAdmin(object):
     delete_recursive = True
 
     filter_mapping = FilterMapping
-    filter_converter = FilterModelConverter
+    filter_converter = AdminFilterModelConverter
 
     # templates, to override see get_template_overrides()
     base_templates = {
@@ -103,7 +116,7 @@ class ModelAdmin(object):
     def get_filter_form(self):
         return FilterForm(
             self.model,
-            self.filter_converter(),
+            self.filter_converter(self),
             self.filter_mapping(),
             self.filter_fields,
             self.filter_exclude,
@@ -126,7 +139,7 @@ class ModelAdmin(object):
     def get_add_form(self):
         return self.get_form(adding=True)
 
-    def get_edit_form(self):
+    def get_edit_form(self, instance):
         return self.get_form()
 
     def get_query(self):
@@ -170,6 +183,9 @@ class ModelAdmin(object):
                 query = query.order_by(field.asc() if not desc else field.desc())
         return query
 
+    def get_extra_context(self):
+        return {}
+
     def index(self):
         query = self.get_query()
 
@@ -196,6 +212,7 @@ class ModelAdmin(object):
             filter_form=filter_form,
             field_tree=field_tree,
             active_filters=cleaned,
+            **self.get_extra_context()
         )
 
     def dispatch_save_redirect(self, instance):
@@ -210,17 +227,23 @@ class ModelAdmin(object):
 
     def add(self):
         Form = self.get_add_form()
+        instance = self.model()
 
         if request.method == 'POST':
             form = Form(request.form)
             if form.validate():
-                instance = self.save_model(self.model(), form, True)
+                instance = self.save_model(instance, form, True)
                 flash('New %s saved successfully' % self.get_display_name(), 'success')
                 return self.dispatch_save_redirect(instance)
         else:
             form = Form()
 
-        return render_template(self.templates['add'], model_admin=self, form=form)
+        return render_template(self.templates['add'],
+            model_admin=self,
+            form=form,
+            instance=instance,
+            **self.get_extra_context()
+        )
 
     def edit(self, pk):
         try:
@@ -228,7 +251,7 @@ class ModelAdmin(object):
         except self.model.DoesNotExist:
             abort(404)
 
-        Form = self.get_edit_form()
+        Form = self.get_edit_form(instance)
 
         if request.method == 'POST':
             form = Form(request.form, obj=instance)
@@ -239,7 +262,12 @@ class ModelAdmin(object):
         else:
             form = Form(obj=instance)
 
-        return render_template(self.templates['edit'], model_admin=self, instance=instance, form=form)
+        return render_template(self.templates['edit'],
+            model_admin=self,
+            instance=instance,
+            form=form,
+            **self.get_extra_context()
+        )
 
     def collect_objects(self, obj):
         deps = obj.dependencies()
@@ -280,13 +308,16 @@ class ModelAdmin(object):
             model_admin=self,
             query=query,
             collected=collected,
+            **self.get_extra_context()
         ))
 
-    def collect_related_fields(self, model, accum, path):
+    def collect_related_fields(self, model, accum, path, seen=None):
+        seen = seen or set()
         path_str = '__'.join(path)
         for field in model._meta.get_fields():
-            if isinstance(field, ForeignKeyField):
-                self.collect_related_fields(field.rel_model, accum, path + [field.name])
+            if isinstance(field, ForeignKeyField) and field not in seen:
+                seen.add(field)
+                self.collect_related_fields(field.rel_model, accum, path + [field.name], seen)
             elif model != self.model:
                 accum.setdefault((model, path_str), [])
                 accum[(model, path_str)].append(field)
@@ -322,6 +353,7 @@ class ModelAdmin(object):
             active_filters=cleaned,
             related_fields=related,
             sql=query.sql(self.db.get_compiler()),
+            **self.get_extra_context()
         )
 
     def ajax_list(self):
@@ -421,7 +453,7 @@ class AdminTemplateHelper(object):
             return field.verbose_name
 
     def get_model_admins(self):
-        return {'model_admins': self.admin.get_model_admins()}
+        return {'model_admins': self.admin.get_model_admins(), 'branding': self.admin.branding}
 
     def get_admin_url(self, obj):
         model_admin = self.admin.get_admin_for(type(obj))
@@ -457,7 +489,7 @@ class AdminTemplateHelper(object):
 
 class Admin(object):
     def __init__(self, app, auth, template_helper=AdminTemplateHelper,
-                 prefix='/admin', name='admin'):
+                 prefix='/admin', name='admin', branding='flask-peewee'):
         self.app = app
         self.auth = auth
 
@@ -471,6 +503,8 @@ class Admin(object):
         self.template_helper = template_helper(self)
         self.template_helper.prepare_environment()
 
+        self.branding = branding
+
     def auth_required(self, func):
         @functools.wraps(func)
         def inner(*args, **kwargs):
